package CompilerError;

import java.io.*;
import java.util.*;
import MyParser;
import MyLexer;

public class ErrorMsg {

  public ErrorMsg( PrintWriter out ) {
    this.out = out;
    showNameLineColumn = true;
    wasAnCompilerError = false;
    classTable = new Hashtable();
  }
  
  private Hashtable classTable;
  
  public Hashtable getClassTable() {
    return classTable;
  }

  public void setLexer( MyLexer lexer, String fileName ) {
    this.lexer = lexer;
    this.fileName = fileName;
  }

  public void setShowNameLineColumn(boolean showNameLineColumn) {
     this.showNameLineColumn = showNameLineColumn;
  }

  public void show(String msg, int lineNumber, int columnNumber ) {

     if ( showNameLineColumn ) {
        out.print(fileName);
        out.print(":line(" + lineNumber + ") Column(" +
           columnNumber + "): ");
     }
     out.println(msg);
     if ( showNameLineColumn ) {
        out.println(lexer.getCurrentLine());
     }
     wasAnCompilerError = true;
     out.flush();
     throw new CompilerException();
  }

  public void show(String msg, String className, int lineNumber ) {
    
      Vector errorArray = (Vector ) classTable.get(className);
      if ( errorArray == null ) {
         errorArray = new Vector();
         classTable.put( className, errorArray );
      }
      errorArray.addElement( new CompilerError(msg, lineNumber) );
  }

  public void show(String msg ) {

     if ( showNameLineColumn ) 
        show(msg, lexer.getLineNumber(), lexer.getColumnNumber() );
     else
        show(msg, 0, 0);

  }

  public void signal(String msg ) {

     show(msg);
     throw new CompilerException();
  }

  public boolean getWasAnCompilerError() {
    return wasAnCompilerError;
  }

  public static void init( PrintWriter p_staticOut ) {
    staticOut = p_staticOut;
   }

  public static void signalFatalError( String msg ) {
     if ( staticOut != null )
        staticOut.println(msg);
     else {
        System.out.println("Fatal error and compiler internal error: staticOut not initialized in ErrorMsg");
        System.out.flush();
     }
  }

  private static PrintWriter staticOut;

  private MyLexer lexer;
  private PrintWriter out;
  private boolean wasAnCompilerError;
  private String fileName;
  private boolean showNameLineColumn;
}

