class DQueue( T : Any ) subclassOf Container(T)
     /* A queue of elements.
        This class is better when the number of list elements 
        varies greatly during the container lifetime. */
  
   proc init()
      begin
      head = tail = nil;
      size = 0;
      end
  
   public:
  
      proc add( elem : T )
                     ( exception : CatchOutOfMemoryException )
         begin  
         if head == nil
         then
            head = ListNo.new(elem);
            tail = head;
         else
            var no : ListNo = ListNo.new(elem);
            no.setSucc(nil);
            head.setSucc(no);
            head = no;
         endif
         ++size;
         end
         
         
      proc get()
              ( exception : CatchNotFoundException ) : T
         begin
         if tail == nil
         then
            exception.throw( NotFoundException.new() );
         else
            return tail.getValue();
         endif
         end
                              
      proc remove()
                 ( exception : CatchNotFoundException ) : T
         begin
         if tail == nil
         then
            exception.throw( NotFoundException.new() );
         else
            var x : T = tail.getValue();
            tail = tail.getSucc();
            if tail == nil
            then
               head = nil;
            endif
            --size;
            return x;
         endif
         end
                              

      proc removeAll()
         // empties the container
         begin
         head = tail = nil;
         size = 0;
         end
                                                      
      proc getSize() : integer
         // returns the number of elements in the container
         begin
         return size;
         end
      
      proc empty() : boolean
         // returns true if there is no element in the container
         begin
         return first == nil;
         end
      
      proc full() : boolean
            /* returns true if there is no more space in the container. This
               method only returns true in static containers whose number of
               elements is defined at the creation of the container */
         begin
         return false;
         end
      
      proc getIter() : Iter(T)
            // returns an iterator for the container
         begin
         return DListIter.new(first, size);
         end
      
      proc forEach( f : Function(T) )
            // calls "f.exec(x)" for each element x of the container
         begin
         var no : ListNo(T) = tail;
         while no <> nil do
            begin
            f.exec(no.getValue());
            no = no.getSucc();
            end
         end


      proc collect( f : Filter(T) ) : Container(T)
         /* collects all container elements x such that "f.test(x)"
            evaluates to true. These elements are inserted in a new
            container that is returned by this method. */
            var dqueue : DQueue(T);
         begin
         dqueue#init();
         var no : ListNo(T) = tail;
         while no <> nil do
            begin
            if f.test(no.getValue())
            then
               dqueue.add(no.getValue());
            endif
            no = no.getSucc();
            end
         return dqueue;
         end
         
         
      proc reset()
         begin
         current = tail;
         end
  
      proc next() : T
         begin
         if current == nil
         then
            return nil;
         else
            var x : T = current.getValue();
            current = current.getSucc();
            return x;
         endif
         end
      
      proc endIter()
      begin
      end

   private:
   
      var current : ListNo;
          size : integer;
          head, tail : ListNo;
end

        
