package Interface;

import AST.*;
import java.util.*;
import java.io.*;
import CompilerError.*;
import NameServer;
import ws;


public class ClassObjectTypeInterface extends TypeInterface {

    public ClassObjectTypeInterface() {
        super();
        initMethod = null;
        publicConstList = new ConstList();
        privateConstList = new ConstList();
        covl = new ClassObjectVariableList();
    }
    
    
   public void genI( PrintWriter out ) {
      out.println("object " + name);
      out.println();
      if ( initMethod != null ) {
         initMethod.genI(out);
         out.println();
      }
      
      if ( publicMethodList.getSize() != 0 || publicConstList.getSize() != 0 ) {
         out.println("   public:");
         
         if ( publicMethodList.getSize() != 0 ) {
            publicMethodList.genI(out);
            out.println();
         }
         if ( publicConstList.getSize() != 0 ) {
            publicConstList.genI(out);
            out.println();
         }
      }
      
      if ( privateMethodList.getSize() != 0 || covl.getSize() != 0 
           || privateConstList.getSize() != 0 ) {
            
         out.println("   private:");
         if ( privateConstList.getSize() != 0 ) {
            privateConstList.genI(out);
            out.println();
         }
         if ( privateMethodList.getSize() != 0 ) {
            privateMethodList.genI(out);
            out.println();
         }
         if ( covl.getSize() != 0 ) {
            covl.genI(out);
            out.println();
         }
      }
      out.println("end");
   }

    
    public void reify( TypeInterfaceList til, ErrorMsg em ) {

        super.reify(til, em);
        if ( initMethod != null ) 
           initMethod.reify(til, em);
        publicConstList.reify(til, em);
        privateConstList.reify(til, em);
        covl.reify(til, em);
    }
    
    
    
    
    public void setInitMethod( MethodInterface mi ) {
        initMethod = mi;
    }
    
    public MethodInterface getInitMethod() {
        return initMethod;
    }
    
    public Constant searchConstant( String name ) {
        Enumeration e = publicConstList.elements();
        
        while ( e.hasMoreElements() ) {
            Constant constant = (Constant ) e.nextElement();
            if ( constant.getName().compareTo(name) == 0 )
               return constant;
        }
        return null;
    }
    
    public ConstList getPublicConstList() {
        return publicConstList;
    }

    public ConstList getPrivateConstList() {
        return privateConstList;
    }
    
    public void setPublicConstList( ConstList constList ) {
        publicConstList = constList;
    }
    
    public void setPrivateConstList( ConstList constList ) {
        privateConstList = constList;
    }
    
    public void setClassObjectVariableList( ClassObjectVariableList covl ) {
        this.covl = covl;
    }
    
    public ClassObjectVariableList getClassObjectVariableList() {
        return covl;
    }
    
    public ClassObjectType getClassObjectType() {
        return classObjectType;
    }
    
    public MethodInterface fullSearch( String methodSearchName ) {
        MethodInterface mi;
        
        if ( initMethod != null && initMethod.getSearchName().compareTo(methodSearchName) == 0 )
           return initMethod;
        else
           if ( (mi = getPublicMethodList().get(methodSearchName)) != null ) 
              return mi;
           else
              return getPrivateMethodList().get(methodSearchName);
    }

    
    public Hashtable multipleFullSearch( String anySearchMethodName ) {
        Hashtable v = new Hashtable();
        
        if ( initMethod != null && 
             initMethod.getAnySearchName().compareTo(anySearchMethodName) == 0 ) 
            v.put(initMethod.getSearchName(), initMethod);
        getPublicMethodList().multipleGet(anySearchMethodName, v);
        getPrivateMethodList().multipleGet(anySearchMethodName, v);
        return v;
    }
        
    
    
    public Constant fullSearchConstant( String name ) {
        Enumeration e = publicConstList.elements();
        Constant constant;
        
        while ( e.hasMoreElements() ) {
            constant = (Constant ) e.nextElement();
            if ( constant.getName().compareTo(name) == 0 )
               return constant;
        }
        e = privateConstList.elements();
        
        while ( e.hasMoreElements() ) {
            constant = (Constant ) e.nextElement();
            if ( constant.getName().compareTo(name) == 0 )
               return constant;
        }

        return null;
    }

    
    public void setName( String name ) {
       super.setName(name);
       classObjectType = new ClassObjectType(this);
    }
    
    
    public String getJavaName() {
       if ( javaName == null ) {
          javaName = TypeInterface.buildClassObjectName(name);
       }
       return javaName;
    }

    public ClassTypeInterface getClassTypeInterface() {
       return classTypeInterface;
    }

    public void setClassTypeInterface( ClassTypeInterface classTypeInterface ) {
       this.classTypeInterface = classTypeInterface;
    }

    public String getAccessToClassObject() {
        return NameServer.getAccessToClassObject(classTypeInterface.getName());
    }
    
    public String getAccessToClassInfo() {
       return NameServer.getAccessToClassInfoClassObject( getJavaName() );
    }
    
    public void genInterfaceNameList( PrintWriter out ) {

       boolean first = true;
       Hashtable h = new Hashtable();
       TypeInterface current = this;
       while ( true ) {
          MethodInterfaceList mil = current.getPublicMethodList();
          Enumeration e = mil.anyElements();
          while ( e.hasMoreElements() ) {
             String interfaceName = ((MethodInterface ) e.nextElement()).getMethodInterfaceName();
             if ( h.put(interfaceName, "") == null ) {
                if ( ! first ) 
                   out.println(",");
                ws.print( interfaceName );
                first = false;
             }
          }
          if ( current == Type.anyClassObjectTypeInterface )
             break;
          else {
             current = Type.anyClassObjectTypeInterface;
          }
       }
        
     }
    
    
    
    private ClassObjectType classObjectType;
    private MethodInterface initMethod;
    private ConstList publicConstList, privateConstList;
    private ClassObjectVariableList covl;
    private ClassTypeInterface classTypeInterface;

}
