package Interface;

import java.util.*;
import AST.*;
import java.io.*;
import WhiteSpace;
import CompilerError.*;
import NameServer;
import ws;
import GreenProject;

public class MethodInterface {
    
    public MethodInterface( String name, boolean isAbstract, int methodNumber ) {
        this.name = name;
        this.isAbstract = isAbstract;
        this.methodNumber = methodNumber;
        exceptionType = null;
        searchName = null;
    }
 
    public void genI( PrintWriter out ) {
        WhiteSpace.printMember(out);
        if ( isAbstract ) 
           out.print("abstract ");
        out.print("proc " + name + "(");
        if ( paramList.getSize() > 0 )
           out.print(" ");
        paramList.genI(out, name.length());
        if ( paramList.getSize() > 0 )
           out.print(" ");
        out.print(")");
        if ( exceptionType != null ) {
           out.println();
           WhiteSpace.printStat(out);
           out.print("( exception : " + exceptionType.getGreenName() + " )");
        }
        if ( returnType != null )
           out.print( " : " + returnType.getGreenName() );
        out.println();
    }
        
    
    public void genNewMethod( PrintWriter out, String className ) {
        out.print("proc new(");
        int size = paramList.getSize();
        Enumeration f = paramList.elements();
        while ( f.hasMoreElements() ) {
           Parameter p = (Parameter ) f.nextElement();
           out.print( p.getName() + " : " + p.getType().getGreenName() );
           if ( --size > 0 )
              out.print("; ");
        }
        out.println(") : " + className);
    }
    
    public void reify( TypeInterfaceList il, ErrorMsg errorMsg ) {
    
       ClassTypeInterface cti;
       
       if ( returnType != null )
          if ( (cti = il.get(returnType.getInterfaceName())) == null ) 
             errorMsg.show("Type " + returnType.getInterfaceName() + " was not found");
          else
             returnType.setTypeInterface(cti);
       if ( paramList != null ) 
          paramList.reify(il, errorMsg);
       if ( exceptionType != null ) 
          if ( (cti = il.get(exceptionType.getInterfaceName())) == null ) 
             errorMsg.show("Type " + exceptionType.getInterfaceName() + " was not found");
          else
             exceptionType.setTypeInterface(cti);
       if ( localVariableList != null ) 
          localVariableList.reify(il, errorMsg);
    }
    

 
    public void setParamList( ParameterList paramList ) {
        this.paramList = paramList;
    }
    
    public ParameterList getParamList() {
        return paramList;
    }
    
    
    public String getName() {
        // name of the method. Not the searchName
        return name;
    }
    
    public String getSearchName() {
        
        if ( searchName == null ) {
           searchName = TypeInterface.buildMethodSearchName(name, paramList);
        }
        return searchName;
    }
     
     
    public String getAnySearchName() {
        
        if ( anySearchName == null ) {
           anySearchName = TypeInterface.buildAnySearchName(name, paramList);
        }
        return anySearchName;
    }
     
     
           
    public boolean getIsAbstract() {
        return isAbstract;
    }
    
    public void setIsAbstract( boolean isAbstract ) {
        this.isAbstract = isAbstract;
    }
    
    public void setVisibility( int visibility ) {
        this.visibility = visibility;
    }
    
    public int getVisibility() {
        return visibility;
    }
           
           
    public void setExceptionType( Type exceptionType ) {
        this.exceptionType = exceptionType;
    }
    
    
    public Type getExceptionType() {
        return exceptionType;
    }
    
    public void setReturnType( Type returnType ) {
        this.returnType = returnType;
    }
    

    public Type getReturnType() {
        return returnType;
    }
    
    public int getMethodNumber() {
        return methodNumber;
    }
    
    public void setLocalVariableList( LocalVariableList localVariableList ) {
        this.localVariableList = localVariableList;
    }
    
    public LocalVariableList getLocalVariableList() {
        return localVariableList;
    }
    
    public String getJavaName() {
       if ( javaName == null ) {
          javaName = NameServer.getMethodJavaName(name, paramList);
       }
       return javaName;
    }
    
    public String getMethodInterfaceName() {
       if ( methodInterfaceName == null ) {
          methodInterfaceName = NameServer.getMethodInterfaceName( 
             name, returnType, paramList );
       }
       return methodInterfaceName;
    }

    
    public void genForInterceptAll( PrintWriter out, String attachedJavaClassName ) {
       ws.add();
       ws.print( "public " );
       if ( returnType != null ) {
          if ( returnType instanceof BasicType || 
               returnType.getIsJavaClass() ||
               GreenProject.subtypeEqualSubclass || 
               returnType.isArrayType()  
               )
             out.print( returnType.getJavaName() + " " );
          else 
             out.print( "_Any " );
       }
       else
          out.print("void ");
      
       out.print( getJavaName() + "(" );
       paramList.gen(out);
       out.println(") {");
       ws.add();
       ws.println("$Array$1$_Any args = new $Array$1$_Any(");
       ws.add();
       ws.println("new _Any[] {");
       ws.add();
       Enumeration e = paramList.elements();
       int size = paramList.getSize();
       while ( e.hasMoreElements() ) {
          Parameter p = (Parameter ) e.nextElement();
          Type t = p.getType();
          if ( t instanceof BasicType ) {
             ws.print( ((BasicType ) t).getNewWrapperString() + p.getJavaName() + ")" );
          }
          else if ( t.getName().compareTo("String") == 0 )
             ws.print("(_PackedString ) _PackedString.$classObject._new$String(" + p.getJavaName() + ")");
          else
             ws.print(p.getJavaName());
          if ( --size > 0 )
             out.println(",");
          else
             out.println();
       }
       
       ws.println("}");
       ws.sub();
       ws.println(");");
       ws.sub();

       String value = "this._interceptAll$_ObjectMethodInfo$$Array$1$_Any(" +  
         attachedJavaClassName + ". " + 
         NameServer.getObjectMethodInfoNameVar(name, paramList) + ", args)";
       if ( returnType == null ) 
          ws.println( value + ";" );
       else {
          ws.print("return ");
          if ( returnType instanceof BasicType ) {
             out.print( "((" + 
                 ((BasicType ) returnType).getJavaWrapperName() + " ) " + value
                 + ")._get()" );
          }
          else
             if ( returnType.getName().compareTo("String") == 0 ) {
                out.print( "((_PackedString ) " + value + ")._getString()");
             }
             else
                out.print( value );
         out.println(";");
       }
       ws.sub();
       ws.println("}");
       ws.sub();
       out.println();
    }
    
    private String name;
    private String javaName;
    private String methodInterfaceName;
    private Type returnType;
    private ParameterList paramList;
    private String searchName, anySearchName;
    private boolean isAbstract;
    private int visibility;
    
    private Type exceptionType;
    private int methodNumber;
    private LocalVariableList localVariableList;
    
}
