package Interface;

import CompilerError.*;
import java.util.*;
import java.io.*;
import AST.*;
import GreenProject;

/*
   TypeInterface may represent a class or class object. 
   If this interface represents a class object, classObjectInterface is null.
*/

abstract public class TypeInterface {

    public TypeInterface() {
        name = null;
        privateMethodList = new MethodInterfaceList();
        publicMethodList = new MethodInterfaceList();
        isBasicType = false;
        isJavaClass = false;
    }

    
    abstract public void genI( PrintWriter out );
    
    public void reify( TypeInterfaceList til, ErrorMsg errorMsg ) {

       if ( privateMethodList != null )
          privateMethodList.reify(til, errorMsg);
       if ( publicMethodList != null )
          publicMethodList.reify(til, errorMsg);
        
    }
    

    public boolean isSupertypeOf( TypeInterface other ) {
       
       if ( other.getName().compareTo("Nil") == 0 ||
            name.compareTo(other.getName()) == 0 )
            return true;
       Enumeration e = publicMethodList.elements();
       while ( e.hasMoreElements() ) {
          MethodInterface m = publicMethodList.get( (String ) e.nextElement());
          if ( other.publicSearch( m.getSearchName() ) == null ) 
             return false;
       }
       return true;
    }

    abstract public void genInterfaceNameList( PrintWriter out );
    
    
    public MethodInterface publicSearch( String methodName, ExprList exprList ) {
       // typeArray has types with searchNames
       
       return publicSearch(TypeInterface.buildMethodSearchName(methodName, exprList));
    }
    

    
    public MethodInterface publicSearch( String methodSearchName ) {
       return publicMethodList.get(methodSearchName);
    }
    

    public Hashtable multiplePublicSearch( String methodName, ExprList exprList ) {
        return multiplePublicSearch( TypeInterface.buildAnySearchName(methodName, exprList) ); 
    }
       
    public Hashtable multiplePublicSearch( String anySearchMethodName ) {
       /* return a Vector with all methods of the interface with name anySearchMethodName.
          */
       return multiplePublicSearch(anySearchMethodName, new Hashtable());
    }
       
    public Hashtable multiplePublicSearch( String anySearchMethodName, Hashtable result ) {
       return publicMethodList.multipleGet(anySearchMethodName, result);
    }
        
          
    public String getName() {
        // return the name of the class or class object of the interface. 
        // Probably this method will not be used
        return name;
    }
    
    public void setName( String name ) {
        this.name = name;
    }
    
    public void setPublicMethodList( MethodInterfaceList publicMethodList ) {
        this.publicMethodList = publicMethodList;
    }
    
    public MethodInterfaceList getPublicMethodList() {
        return publicMethodList;
    }
    
    public void setPrivateMethodList( MethodInterfaceList privateMethodList ) {
        this.privateMethodList = privateMethodList;
    }
    
    public MethodInterfaceList getPrivateMethodList() {
        return privateMethodList;
    }
    
    public Constant searchConstant( String name ) {
        return null;
    }
    
    
    public MethodInterface fullSearch( String methodName, ExprList exprList ) {
        return fullSearch( buildMethodSearchName(methodName, exprList) );
    }
    
    
    public Hashtable multipleFullSearch( String methodName, ExprList exprList ) {
        return multipleFullSearch( TypeInterface.buildAnySearchName(methodName, exprList) );
    }
    
    abstract public Hashtable multipleFullSearch( String anySearchMethodName );
    
    
    public static String buildMethodSearchName(String methodName, ExprList exprList ) {
        
        String methodSearchName = methodName + "$" + exprList.getSize();
        Enumeration e = exprList.elements();
        while ( e.hasMoreElements() ) {
            Expr expr = (Expr ) e.nextElement();
            methodSearchName = methodSearchName + "$" + expr.getType().getJavaName();
        }
        return methodSearchName;
    }
    

    public static String buildMethodSearchName(String methodName, ParameterList paramList) {
        
        String methodSearchName = methodName + "$" + 
            paramList.getSize() + buildParameterListName(paramList);
        return methodSearchName;
    }
    
    public static String buildParameterListName( ParameterList paramList ) {
        String paramListName = "";
        Enumeration e = paramList.elements();
        while ( e.hasMoreElements() ) {
            Parameter p = (Parameter ) e.nextElement();
            paramListName = paramListName + "$" + p.getType().getJavaName();
        }
        return paramListName;
    }


    public static String buildAnySearchName(String methodName, ExprList exprList ) {
        // Put Any in place of the names of classes. So  "proc m(real, Person)"
        // results in  m$2$float$_Any
        
        String methodSearchName = methodName + "$" + exprList.getSize();
        Enumeration e = exprList.elements();
        while ( e.hasMoreElements() ) {
            Expr expr = (Expr ) e.nextElement();
            if ( expr.getType() instanceof BasicType ||
                 GreenProject.subtypeEqualSubclass ||
                 expr.getType().getIsJavaClass() ) 
               methodSearchName = methodSearchName + "$" + expr.getType().getJavaName();
            else
               methodSearchName = methodSearchName + "$_Any";
        }
        return methodSearchName;
    }
    

    public static String buildAnySearchName(String methodName, ParameterList paramList) {
        // Put Any in place of the names of classes. So  "proc m(real, Person)"
        // results in  m$2$float$_Any
        
        String methodSearchName = methodName + "$" + paramList.getSize() +
           buildAnyParameterListName(paramList);
        /*
        Enumeration e = paramList.elements();
        while ( e.hasMoreElements() ) {
            Parameter p = (Parameter ) e.nextElement();
            if ( p.getType() instanceof BasicType ) 
               methodSearchName = methodSearchName + "$" + p.getType().getJavaName();
            else
               methodSearchName = methodSearchName + "$_Any";
        }
        */
        return methodSearchName;
    }

    public static String buildAnyParameterListName( ParameterList paramList ) {
        String s = "";
        Enumeration e = paramList.elements();
        while ( e.hasMoreElements() ) {
            Parameter p = (Parameter ) e.nextElement();
            if ( p.getType() instanceof BasicType || 
                 GreenProject.subtypeEqualSubclass ||
                 p.getType().getIsJavaClass() ) 
               s = s + "$" + p.getType().getJavaName();
            else
               s = s + "$_Any";
        }
        return s;
    }

    public void setIsJavaClass( boolean isJavaClass ) {
       this.isJavaClass = isJavaClass;
    }
    
    public boolean getIsJavaClass() {
       return isJavaClass;
    }
    
    abstract public MethodInterface fullSearch(String methodSearchName);
    
    
    public static String buildClassName( String className, ExprList exprList ) {
        if ( exprList != null && exprList.getSize() > 0 ) {
           className = "$pc$" + className + "$" + exprList.getSize();
           Enumeration e = exprList.elements();
           while ( e.hasMoreElements() ) {
               Expr expr = (Expr ) e.nextElement();
               if ( ! (expr instanceof ClassObjectExpr) )
                  return null;
               Type t = ((ClassObjectExpr ) expr).getClassObjectType();
               if ( ! (t.getTypeInterface() instanceof ClassObjectTypeInterface) ) 
                  return null;
               ClassObjectTypeInterface coti = (ClassObjectTypeInterface ) t.getTypeInterface();
               className = className + "$" + coti.getClassTypeInterface().getJavaName();
           }
        }
        return className;
    }
    

    public static String buildClassName( String className, TypeList typeList ) {
        if ( typeList != null && typeList.getSize() > 0 ) {
           className = "$pc$" + className + "$" + typeList.getSize();
           Enumeration e = typeList.elements();
           while ( e.hasMoreElements() ) 
               className = className + "$" + ((Type ) e.nextElement()).getJavaName();
        }
            
        return className;
    }
    
    public static String buildClassObjectName( String className ) {
        if ( basicTypes.get(className) != null ) 
           return "$co$_" + className;
        else
           return "$co$" + className;
    }
    
    public static String buildClassObjectName( TypeInterface cti ) {
        return buildClassObjectName(cti.getName());
    }
    
    public static String buildArrayName( Type type, int dim ) {
        return "$Array$" + dim + "$" + type.getJavaName();
    }
    
    public static String buildArrayName( String javaName, int dim ) {
        return "$Array$" + dim + "$" + javaName;
    }
    
    public static String buildJavaName( String name ) {
        
       char ch = name.charAt(0);
       if ( ch == '$' || ch == '_' )
          return name;
       else {
          String s;
          if ( (s = (String ) basicClassTable.get(name)) != null ) 
             return s;
          else 
             return "_" + name;
       }
    }
    
    public static String buildRealClassJavaName( String name ) {
        if ( basicTypes.get(name) == null )
           return buildJavaName(name);
        else
           return "__" + name;
    }
        
    
    public static String buildBasicTypeClassName( String name ) {
       if ( name.compareTo("String") == 0 )
          return "_"  + name;
       else 
          return "__"  + name;
    }


    static public Hashtable basicClassTable, basicTypes;
    static {
        basicClassTable = new Hashtable();
        basicClassTable.put("char", "char");
        basicClassTable.put("byte", "byte");
        basicClassTable.put("boolean", "boolean");
        basicClassTable.put("integer", "int");
        basicClassTable.put("long", "long");
        basicClassTable.put("real", "float");
        basicClassTable.put("double", "double");
        basicClassTable.put("String", "String");

        basicTypes = new Hashtable();
        basicTypes.put("char", "");
        basicTypes.put("byte", "");
        basicTypes.put("boolean", "");
        basicTypes.put("integer", "");
        basicTypes.put("long", "");
        basicTypes.put("real", "");
        basicTypes.put("double", "");
    
    }
    
    
    
    public Constant fullSearchConstant( String name ) {
        return null;
    }
    
    
    abstract public String getJavaName();
    
    public String getInterfaceName() {
       return getJavaName();
    }

    
    public void setToAddTextPublic( char []publicText ) {
        this.publicText = publicText;
    }
    
    public char []getToAddTextPublic() {
        return publicText;
    }

    public void setToAddTextPrivate( char []privateText ) {
        this.privateText = privateText;
    }
    
    public char []getToAddTextPrivate() {
        return privateText;
    }
    
    abstract public String getAccessToClassInfo();

    public void setIsBasicType( boolean isBasicType ) {
       this.isBasicType = isBasicType;
    }
    
    public boolean getIsBasicType() {
       return isBasicType;
    }
    
    private boolean isBasicType;
    private char []publicText;
    private char []privateText;
    
    protected MethodInterfaceList privateMethodList;
    protected String name;
    protected String javaName;
      // public methods
    protected MethodInterfaceList publicMethodList;
    
    protected boolean isJavaClass;
}
    


    