import AST.*;
import Interface.*;
import java.util.*;

public class NameServer {
    
   public static String getClassMethodInfoNameVar( String methodName, ParameterList paramList ) {
      return "$cmi$" + methodName + TypeInterface.buildParameterListName(paramList);
   }
   
   public static String getObjectMethodInfoNameVar( String methodName, ParameterList paramList ) {
      return "$omi$" + methodName + TypeInterface.buildParameterListName(paramList);
   }
   
   public static String getMethodJavaName( String methodName, ParameterList paramList ) {

      char ch = methodName.charAt(0);
      if ( Character.isLetter(ch) || ch == '$' || ch == '_' )
         return "_" + methodName + TypeInterface.buildParameterListName(paramList);
      else 
         // should be an operator
         return NameServer.getOperatorName(methodName);
    
   }
   
   public static String getMethodInterfaceName( String methodName, Type returnType,
          ParameterList paramList ) {
            
      char ch = methodName.charAt(0);
      if ( ! Character.isLetter(ch) && ch != '$' && ch != '_' )
         methodName = NameServer.getOperatorName(methodName);
            
      String name = "$mi$" + methodName + "$" + paramList.getSize() + 
          TypeInterface.buildParameterListName(paramList); 
      if ( returnType != null ) {
//          if ( returnType instanceof BasicType || GreenProject.subtypeEqualSubclass )
             name = name + "$" + returnType.getJavaName();
//          else 
//             name = name + "$_Any";
      }
      return name;
   }
   
   public static String getParameterNameVar( String paramName ) {
      return "$p$" + paramName;
   }
   
   public static String getOperatorName( String stringOperator ) {
       return (String ) operatorNameTable.get(stringOperator);
   }
 
   public static String getAccessToClassObject( String name) {
      return TypeInterface.buildRealClassJavaName(name) + ".$classObject";
   }
   
   public static String getAccessToClassInfoClass( String javaName ) {
      return javaName + ".$i";
   }
   
   public static String getAccessToClassInfoClassObject( String javaName ) {
      return javaName + ".$i";
   }
 
   public static String getShellClassName( String shellName, ClassType attached ) {
      return "$shell$" + TypeInterface.buildJavaName(shellName) + "$" + 
         attached.getJavaName();
   }
   
   public static String getShell_ivClassName( String shellJavaName ) {
      return "$shellivc$" + shellJavaName;
   }
   
   public static String getJavaFileName(String name, String javaName) {
      if ( TypeInterface.basicTypes.get(name) == null ) 
         return javaName;
      else
         return "__" + name;
   }
   
   public static Hashtable operatorNameTable;   
   static {
      operatorNameTable = new Hashtable();
      operatorNameTable.put("and", "and");
      operatorNameTable.put("not", "not");
      operatorNameTable.put("or", "or");
      operatorNameTable.put("xor" , "xor");
      operatorNameTable.put("&", "binand");
      operatorNameTable.put("~", "binnot");
      operatorNameTable.put("|", "binor");
      operatorNameTable.put("^" , "binxor");
      operatorNameTable.put(">>", "rightshift");
      operatorNameTable.put("<<", "leftshift");
      operatorNameTable.put("==", "eq");
      operatorNameTable.put("=" , "assign");
      operatorNameTable.put("<>", "neq");
      operatorNameTable.put(">=", "ge");
      operatorNameTable.put(">" , "gt");
      operatorNameTable.put("<=", "le");
      operatorNameTable.put("<" , "lt");
      operatorNameTable.put("++", "plusplus");
      operatorNameTable.put("--", "minusminus");
      operatorNameTable.put("+" , "plus");
      operatorNameTable.put("-" , "minus");
      operatorNameTable.put("*" , "mult");
      operatorNameTable.put("/" , "div");
      operatorNameTable.put("%" , "remainder");
   }
   
}