package AST;

import ws;
import java.io.*;
import SymbolTable;
import WhiteSpace;
import java.util.*;
import CompilerError.*;
import Interface.*;

public class CaseStatement extends Statement {

    public CaseStatement(Expr expr, Vector eachCaseList, OtherwiseCase otherwiseCase) {
       this.expr = expr;
       this.eachCaseList = eachCaseList;
       this.otherwiseCase = otherwiseCase;
    }

    public void gen( PrintWriter out, Context context ) {
       if ( ! expr.getType().getIsClassObjectType() ) {
         ws.print("switch ( ");
         expr.gen(out, context);
         out.println(" ) {");
         ws.add();
         Enumeration e = eachCaseList.elements();
         while ( e.hasMoreElements() )
            ((EachCase ) e.nextElement()).gen(out, context);
         if ( otherwiseCase != null ) 
            otherwiseCase.gen(out, context);
         ws.sub();
       }
       else {
           // expr is of a class object. Generate if's 
           ws.print("{ " + Type.anyType.getJavaName() + " $any = ");
           expr.gen(out, context);
           out.println(";");
           ws.add();
           int size = eachCaseList.size();
           Enumeration e = eachCaseList.elements();
           while ( e.hasMoreElements() ) {
             size--;
             ((EachCase ) e.nextElement()).genIf(out, context, size == 0);
           }
           if ( otherwiseCase != null )
              otherwiseCase.genIf(out, context);
           ws.println("}");
           ws.sub();
       }
    }

    public void accept( Visitor v ) {
        v.visit(this);
    }

    private Expr expr;
    private Vector eachCaseList;
    private OtherwiseCase otherwiseCase;

}
