package AST;

import java.lang.*;
import ws;
import java.io.*;
import SymbolTable;
import WhiteSpace;
import java.util.*;
import CompilerError.*;
import Interface.*;

public class ExprList {

    public ExprList() {
       exprList = new Vector();
    }

    public void add(Expr expr) {
       exprList.addElement(expr);
    }

    public int getSize() {
        return exprList.size();
    }

    public Enumeration elements() {
       return exprList.elements();
    }

    public void gen( PrintWriter out, Context context ) {
       int size = exprList.size();
       Enumeration e = exprList.elements();
       while ( e.hasMoreElements() ) {
          size--;
          ((Expr ) e.nextElement()).gen(out, context, false);
          if ( size > 0 )
             out.print(", ");
       }
    }

    public String getSearchName() {
        if ( searchName == null ) {
           Enumeration e = exprList.elements();
           searchName = "$" + exprList.size();
           while ( e.hasMoreElements() ) 
              searchName = searchName + "$" + ((Expr ) e.nextElement()).getType().getJavaName();
        }
        return searchName;
    }

    public void check( ErrorMsg em, SymbolTable st, TypeList typeList ) {
        Enumeration typeEnum, exprEnum;
        Type paramType, exprType;
        if ( typeList.getSize() != exprList.size() ) 
           em.show("Wrong number of parameters");
        else {
           typeEnum = typeList.elements();
           exprEnum = exprList.elements();
           while ( typeEnum.hasMoreElements() ) {
               paramType = (Type ) typeEnum.nextElement();
               exprType  = (Type ) exprEnum.nextElement();
               if ( ! paramType.isSupertypeOf(exprType) ) 
                  em.show("Type error");
           }
        }
    }

    public Type getType() {
         if ( exprList == null ) 
            return null;
         else {
            Expr anExpr = (Expr ) exprList.elementAt(0);
            return anExpr.getType();
         }
    }

    public void accept( Visitor v ) {
        v.visit(this);
    }

    private String searchName;
    private Vector exprList;

}
