package AST;

import java.lang.*;
import ws;
import java.io.*;
import SymbolTable;
import WhiteSpace;
import java.util.*;
import CompilerError.*;
import Interface.*;
import CompilerError.*;

abstract public class GenericClass {

    public GenericClass(ClassTypeInterface classTypeInterface) {
       this.classTypeInterface = classTypeInterface;
       shell_ivcName = classTypeInterface.getShell_ivcName();
       superclass = null;
       toAddString = new StringBuffer("");
       javaImportList = new Vector();
       addJavaImport("import java.lang.*;");
       isBasicType = false;
       isShellClass = false;
       
       initMethodList = new InitMethodList();
       publicMethodList = new MethodList();
       subclassMethodList = new MethodList();
       privateMethodList = new MethodList();
       instVarList = new InstanceVariableList();
       
    }

    public void addJavaCode(String toAdd) {
        toAddString.append(toAdd);
    }

    public String getName() {
         return classTypeInterface.getName();
    }

    public boolean getIsAbstract() {
        return false;
    }

    public boolean getIsReflective() {
        return false;
    }

    public void setIsAbstract(boolean isAbstract) {
        ErrorMsg.signalFatalError("Using setIsAbstract with non-NormalClass");
    }

    public void setIsReflective(boolean isReflective) {
        ErrorMsg.signalFatalError("Using setIsReflective with non-NormalClass");
    }

    public void gen( PrintWriter out, boolean useIRL, String className ) {
         
         ws.set(out);
         Enumeration e = javaImportList.elements();
         while ( e.hasMoreElements() ) 
            out.println( (String ) e.nextElement() );
         
         if ( isShellClass )
            out.println("import java.util.*;");
            
         if ( javaImportList.size() > 0 ) {
            out.println();
            out.println();
         }
         
         if ( getIsAbstract() )
            out.print("abstract ");
         out.print("class " + className );
         
         if ( superclass != null ) 
            out.print( " extends " + superclass.getJavaName() );
         else if ( isShellClass )
            out.print( " extends _ReflectiveAnyClass" );
            
         out.println(); 
         if ( ! isBasicType && ! isShellClass ) {
            ws.add();
            ws.println("implements");
            ws.add();
            classTypeInterface.genInterfaceNameList(out);
            out.println(" {");
            ws.sub();
            ws.sub();
         }
         else
            out.println(" {");
         out.println();
         
         if ( isShellClass ) {
            ws.add();
            if ( ((ClassTypeInterface ) attachedType.getTypeInterface()).getIsReflective() ) 
               ws.println("public static Class prev = " + attachedType.getJavaName() +
                  ".class;" );
            else
               ws.println("public static Hashtable ht = new Hashtable();");
            ws.sub();
         }
         
         if ( useIRL && ! isShellClass ) {
            ws.add();
            ws.println("public static " + classTypeInterface.getClassObjectInterface().getJavaName() + 
              " $classObject;");
            ws.println("public static _ClassInfo $i;");
            ws.print("public static ");
            out.print( TypeInterface.buildArrayName("_ObjectInstanceVariableInfo", 1) );
            out.println(" OarrayIVI;");
            ws.print("public static ");
            out.print( TypeInterface.buildArrayName("_ObjectMethodInfo", 1) );
            out.println(" OarrayMI;");
            ws.println("public static _ClassMethodInfo");
            ws.add();
            publicMethodList.genClassMethodInfoVarList(out, 
               subclassMethodList.getSize() > 0 || privateMethodList.getSize() > 0 );
            subclassMethodList.genClassMethodInfoVarList(out, privateMethodList.getSize() > 0);
            privateMethodList.genClassMethodInfoVarList(out, false);
            ws.println(";");
            ws.sub();
            
            if ( initMethodList.getSize() > 0 ) {
               ws.println("public static _ClassInitMethodInfo");
               ws.add();
               initMethodList.genClassMethodInfoVarList(out, false);
               ws.println(";");
               ws.sub();
            }
            
            ws.println("public static _ObjectMethodInfo ");
            ws.add();
            publicMethodList.genObjectMethodInfoVarList(out, 
                subclassMethodList.getSize() > 0 || privateMethodList.getSize() > 0 );
            subclassMethodList.genObjectMethodInfoVarList(out, privateMethodList.getSize() > 0 );
            privateMethodList.genObjectMethodInfoVarList(out, false);
            ws.println(";");
            ws.sub();
            if ( ! isBasicType ) {
               out.println();
               ws.println("public _Any $callMethod( int methodNumber, $Array$1$_Any args ) {");
               ws.add();
               ws.println("_Any any;");
               out.println();
               ws.println("switch (methodNumber) {");
               ws.add();
               initMethodList.genCaseCallMethod(out);
               publicMethodList.genCaseCallMethod(out);
               subclassMethodList.genCaseCallMethod(out);
               privateMethodList.genCaseCallMethod(out);
               ws.println("default :");
               ws.add();
               ws.println("throw new PackedException( _InternalErrorException.$classObject._new$String(\"Error in $callMethod in a class object\") );" );
               ws.sub();
               ws.sub();
               ws.println("}");
               ws.println("return any;");
               ws.sub();
               ws.println("}");
            
               out.println();
               ws.println("void $set( int ivNumber, _Any value ) {");
               ws.add();
               ws.println("switch ( ivNumber ) {");
               ws.add();
               instVarList.genCaseSet(out);
               ws.println("default :");
               ws.add();
               ws.println("throw new PackedException( _InternalErrorException.$classObject._new$String(\"Error in $set in a class object\") );");
               ws.sub();
              
               ws.sub();
               ws.println("}");
               ws.sub();
               ws.println("}");
           
              
               out.println();
               ws.println("_Any $get( int ivNumber ) {");
               ws.add();
               ws.println("switch ( ivNumber ) {");
               ws.add();
               instVarList.genCaseGet(out);
               ws.println("default :");
               ws.add();
               ws.println("throw new PackedException( _InternalErrorException.$classObject._new$String(\"Error in $set in a class object\") );");
               ws.sub();
               ws.sub();
               ws.println("}");
              
               ws.sub();
               ws.println("}");
            }
            
            out.println();
            ws.println("static {");
            ws.add();
            ws.println("$classObject = new " + 
               classTypeInterface.getClassObjectInterface().getJavaName() +
               "();");
            ws.println("$i = new _NormalClassInfo();");
            ws.sub();
            ws.println("}");
            out.println();
       
            ws.println("public static void initialize() {");
            
            ws.add();
            out.println();
            initMethodList.genClassMethodInfo(out);
            publicMethodList.genClassMethodInfo(out);
            subclassMethodList.genClassMethodInfo(out);
            privateMethodList.genClassMethodInfo(out);
            out.println();
            ws.println("((_NormalClassInfo ) $i)._initialize$String$_ClassInfo$$Array$1$_ClassInstanceVariableInfo$$Array$1$_ClassMethodInfo$$Array$1$_ClassInitMethodInfo$_AnyClassObject$boolean$boolean$$Array$1$_CodeAnnotation(");
            ws.add();
            ws.println("\"" + classTypeInterface.getName() + "\"," );
            //changed
            if ( superclass != null ) 
               ws.println( superclass.getTypeInterface().getAccessToClassInfo()  + ",");
            else
               ws.println("null,");
               
            ws.println("new " + 
                TypeInterface.buildArrayName("_ClassInstanceVariableInfo", 1) + "(");
            ws.add();
            ws.println("new _ClassInstanceVariableInfo[] {");
            ws.add();
            instVarList.genVariableInfo(out);
            ws.println("}");
            ws.sub();
            ws.println("),");
            ws.sub();
            
            ws.println("new " + 
                TypeInterface.buildArrayName("_ClassMethodInfo", 1) + "(");
            ws.add();
            ws.println("new _ClassMethodInfo[] {");
            ws.add();
            publicMethodList.genClassMethodInfoVarList(out, 
               subclassMethodList.getSize() > 0 || privateMethodList.getSize() > 0 );
            subclassMethodList.genClassMethodInfoVarList(out, privateMethodList.getSize() > 0 );
            privateMethodList.genClassMethodInfoVarList(out, false);
            ws.println("}");
            ws.sub();
            ws.println("),");
            ws.sub();
            
            ws.println("new " + 
                TypeInterface.buildArrayName("_ClassInitMethodInfo", 1) + "(");
            ws.add();
            ws.println("new _ClassInitMethodInfo[] {");
            ws.add();
            initMethodList.genClassMethodInfoVarList(out, false);
            ws.println("}");
            ws.sub();
            ws.println("),");
            ws.sub();
            
            ws.println("$classObject,");
            ws.println( getIsAbstract() ? "true," : "false," );
            ws.println( getIsReflective() ? "true," : "false," );
            ws.println("null");
            ws.println(");");
            ws.sub();
            
            out.println();
            ws.println( classTypeInterface.getClassObjectInterface().getJavaName() 
                + ".$init($i, $classObject);");
            
            publicMethodList.genObjectMethodInfo(out);
            subclassMethodList.genObjectMethodInfo(out);
            privateMethodList.genObjectMethodInfo(out);
            ws.println("OarrayMI = new " +
                TypeInterface.buildArrayName("_ObjectMethodInfo", 1) + "(");
            ws.add();
            ws.println("new _ObjectMethodInfo[] {");
            ws.add();
            publicMethodList.genObjectMethodInfoList(out, 
               subclassMethodList.getSize() > 0 || privateMethodList.getSize() > 0 );
            subclassMethodList.genObjectMethodInfoList(out, privateMethodList.getSize() > 0);
            privateMethodList.genObjectMethodInfoList(out, false);
            ws.println("}");
            ws.sub();
            ws.println(");");
            ws.sub();
            
            ws.print("OarrayIVI = new ");
            out.println( TypeInterface.buildArrayName("_ObjectInstanceVariableInfo", 1) + "(" );
            ws.add();
            ws.println(" new _ObjectInstanceVariableInfo[] {");
            ws.add();
            instVarList.genObjectInstanceVariableInfo(out);
            ws.println("}");
            ws.sub();
            ws.println(");");
            ws.sub();
            
            
            ws.sub();
            ws.println("}");
            out.println();
            // create a constructor without parameters
            if ( isBasicType || getName().compareTo("String") == 0 )
               ws.print( TypeInterface.buildBasicTypeClassName(getName()) );
            else 
               ws.print(classTypeInterface.getJavaName());
            
            out.println("() { }");
            ws.sub();
         }
         
         Context context = new Context(this, null);
         if ( ! isShellClass ) 
            initMethodList.gen(out, isBasicType, context);
         publicMethodList.gen(out, isBasicType, context);
         subclassMethodList.gen(out, isBasicType, context);
         privateMethodList.gen(out, isBasicType, context);
         if ( ! isShellClass ) 
            instVarList.gen(out);
         
         out.println(toAddString);
         
         if ( classTypeInterface.getHasInterceptAll() ) {
            // create a method in the shell class for each method of the "superclass"
            ClassTypeInterface sc = (ClassTypeInterface ) superclass.getTypeInterface();     
            String attachedJavaClassName = sc.getJavaName();
            Enumeration g = sc.getPublicMethodList().anyElements();
            while ( g.hasMoreElements() ) {
               MethodInterface mi = (MethodInterface ) g.nextElement();
               if ( publicMethodList.search(mi.getSearchName()) == null ) {
                  mi.genForInterceptAll(out, attachedJavaClassName);
               }
            }
         }
         
         ws.println("}");
         ws.sub();
         
         
    }

    public void genShell_ivc(PrintWriter out_ivc) {
        ws.set(out_ivc);
        ws.println("class " + shell_ivcName + " {");
        ws.add();
        ws.println("public static void initialize() { }");
        instVarList.gen(out_ivc, Method.public_v);
        ws.sub();
        ws.println("}");
    }

    public void addJavaImport(String importString) {
        javaImportList.addElement(importString);
    }

    public void genInterfaces( String directory, ErrorMsg errorMsg ) {
         if ( getIsShellClass() )
            return;
         Method m;
         Enumeration e = publicMethodList.elements();
         while ( e.hasMoreElements() ) {
            m = (Method ) e.nextElement();
            if ( ! m.getIsRedefinition() ) {
               String outName = directory + m.getMethodInterfaceName() + ".java";
               try {
                  PrintWriter out = new PrintWriter( 
                      new FileOutputStream(outName) );
                  m.genInterface(out, getName() );
                  out.close();
               }
               catch( Exception excep ) {
                  errorMsg.show("Error writing to file " + outName );
               }
            }
         }
    }

    public ClassTypeInterface getClassTypeInterface() {
       return classTypeInterface;
    }

    public boolean getIsBasicType() {
       return isBasicType;
    }

    public void setIsBasicType( boolean isBasicType ) {
       this.isBasicType = isBasicType;
    }

    public InitMethodList getInitMethodList() {
       return initMethodList;
    }

    public void setInitMethodList( InitMethodList initMethodList ) {
       this.initMethodList = initMethodList;
    }

    public MethodList getPublicMethodList() {
       return publicMethodList;
    }

    public void setPublicMethodList( MethodList publicMethodList ) {
       this.publicMethodList = publicMethodList;
    }

    public MethodList getSubclassMethodList() {
       return subclassMethodList;
    }

    public void setSubclassMethodList( MethodList subclassMethodList ) {
       this.subclassMethodList = subclassMethodList;
    }

    public MethodList getPrivateMethodList() {
       return privateMethodList;
    }

    public void setPrivateMethodList( MethodList privateMethodList ) {
       this.privateMethodList = privateMethodList;
    }

    public InstanceVariableList getInstVarList() {
       return instVarList;
    }

    public void setInstVarList( InstanceVariableList instVarList ) {
       this.instVarList = instVarList;
    }

    public ClassType getSuperclass() {
       return superclass;
    }

    public void setSuperclass( ClassType superclass ) {
       this.superclass = superclass;
    }

    public ClassObject getClassObject() {
       return classObject;
    }

    public void setClassObject( ClassObject classObject ) {
       this.classObject = classObject;
    }

    public boolean getIsShellClass() {
       return isShellClass;
    }

    public void setIsShellClass( boolean isShellClass ) {
       this.isShellClass = isShellClass;
    }

    public ClassType getAttachedType() {
       return attachedType;
    }

    public void setAttachedType( ClassType attachedType ) {
       this.attachedType = attachedType;
    }

    public String getShell_ivcName() {
       return shell_ivcName;
    }

    public void setShell_ivcName( String shell_ivcName ) {
       this.shell_ivcName = shell_ivcName;
    }

    private ClassTypeInterface classTypeInterface;
    private boolean isBasicType;
    private Vector javaImportList;
    private InitMethodList initMethodList;
    private MethodList publicMethodList;
    private MethodList subclassMethodList;
    private MethodList privateMethodList;
    private InstanceVariableList instVarList;
    private ClassType superclass;
    private ClassObject classObject;
    private StringBuffer toAddString;
    private boolean isShellClass;
    private ClassType attachedType;
    private String shell_ivcName;

}
