package AST;

import java.lang.*;
import ws;
import java.io.*;
import SymbolTable;
import WhiteSpace;
import java.util.*;
import CompilerError.*;
import Interface.*;
import GreenProject;

public class InstanceVariable extends Variable {

    public InstanceVariable( String name, int ivNumber ) {
       super(name);
       this.ivNumber = ivNumber;
    }

    public void genDeclaration( PrintWriter out, int visibility ) {
        ws.print( Method.getStringVisibility(visibility) + " " );
        Type t = getType();
        if ( t instanceof BasicType ||
             t.getIsJavaClass() || 
             GreenProject.subtypeEqualSubclass ||
             t.isArrayType()  )
           out.print( t.getJavaName() );
        else
           out.print( "_Any" );
        out.println(" " + getJavaName() + ";");
    }

    public void gen( PrintWriter out, Context context ) {
        if ( context.getCurrentClass() != null &&
             context.getCurrentClass().getIsShellClass() ) 
           out.print( "shellV." + getJavaName() );
        else 
           out.print( "this." + getJavaName() );
    }

    public void genI(PrintWriter out) {
         WhiteSpace.printMember(out);
         out.println("    " + getName() + " : " + getType().getGreenName() + ";");
    }

    public void genCaseSet(PrintWriter out) {
         ws.println("case " + getIvNumber() + " :");
         ws.add();
         ws.println("try {");
         ws.add();
         Type t = getType();
         if ( t instanceof BasicType ) {
            ws.println( getJavaName() + " = ((" + 
                ((BasicType ) t).getJavaWrapperName() + " ) value)._get();" );
         }
         else
            if ( getType().getName().compareTo("String") == 0 ) {
               ws.println( getJavaName() + " = ((_PackedString ) value)._getString();");
            }
            else
               ws.println( getJavaName() + " = (" + getType().getJavaName() + " ) value;" );
            
         
         ws.sub();
         ws.println("} catch ( ClassCastException e ) {");
         ws.add();
         // ws.println("throw new PackedException( _WrongParametersException.$classObject._new() );");
         ws.println("throw new PackedException(e);");
         ws.sub();
         ws.println("}");
         ws.println("break;");
         ws.sub();
    }

    public void genCaseGet(PrintWriter out) {
         ws.println("case " + getIvNumber() + " :");
         ws.add();
         Type t = getType();
         if ( t instanceof BasicType ) {
            ws.println("return " + ((BasicType ) t).getNewWrapperString() + getJavaName() + ");" );
         }
         else if ( t.getName().compareTo("String") == 0 )
            ws.println("return (_PackedString ) _PackedString.$classObject._new$String(" + getJavaName() + ");");
         else
            ws.println("return " + getJavaName() + ";");
         ws.sub();
    }

    public void genVariableInfo(PrintWriter out) {
         ws.println( "(_ClassInstanceVariableInfo ) _ClassInstanceVariableInfo.$classObject._new$String$_ClassInfo$boolean$int$$Array$1$_CodeAnnotation( " );
         ws.add();
         ws.println( "\"" + getName() + "\", " );
         ws.println( ((ClassTypeInterface ) getType().getTypeInterface())
               .getAccessToClassInfo() + "," );
               
         if ( getType() instanceof AtType ) 
            ws.println( "true, " );
         else
            ws.println( "false, " );
         ws.println( getIvNumber() + ", " );
         ws.println("null");
         ws.print(")");
         ws.sub();
    }

    public void genObjectInstanceVariableInfo(PrintWriter out) {
         ws.println( "(_ObjectInstanceVariableInfo ) _ObjectInstanceVariableInfo.$classObject._new$String$_ClassInfo$boolean$int$$Array$1$_CodeAnnotation( " );
         ws.add();
         ws.println( "\"" + getName() + "\", " );
         ws.println( ((ClassTypeInterface ) getType().getTypeInterface())
                .getAccessToClassInfo() + "," );
                
         if ( getType() instanceof AtType ) 
            ws.println( "true, " );
         else
            ws.println( "false, " );
         ws.println( getIvNumber() + ", " );
         ws.println("null");
         ws.print(")");
         ws.sub();
    }

    public int getIvNumber() {
       return ivNumber;
    }

    private int ivNumber;

}
