package AST;

import ws;
import java.io.*;
import SymbolTable;
import WhiteSpace;
import java.util.*;
import CompilerError.*;
import Interface.*;
import GreenProject;

public class MessageSendExpr extends Expr {

    public MessageSendExpr(Expr receiver, MethodInterface methodInterface, ExprList paramList) {
       this.receiver = receiver;
       this.methodInterface = methodInterface;
       this.paramList = paramList;
    }

    public Type getType() {
        return methodInterface.getReturnType();
    }

    public void gen( PrintWriter out, Context context ) {
       //c  maybe print ( and ) around receiver
       Type receiverType = receiver.getType();
       
       if ( receiverType instanceof BasicType ) {
          out.print(  TypeInterface.buildBasicTypeClassName(receiver.getType().getName()) +
                "." + methodInterface.getJavaName() + "(");
          receiver.gen(out, context);
          if ( paramList.getSize() > 0 )
             out.print(", ");
          paramList.gen(out, context);
          out.print(")");
       }
       else {
          if ( receiver instanceof ClassObjectExpr || 
               receiverType.getIsJavaClass() || 
               GreenProject.subtypeEqualSubclass ||
               receiverType.isArrayType()  
               ) {
             receiver.gen(out, context);
             out.print(".");
             if ( receiverType.getIsJavaClass() )
                out.print( methodInterface.getName() );
             else 
                out.print( methodInterface.getJavaName() );
             out.print("(");
             paramList.gen(out, context);
             out.print(")");
          }
          else {
             out.print("((" + methodInterface.getMethodInterfaceName() + " ) ");
             receiver.gen(out, context);
             out.print(")." + methodInterface.getJavaName() + "(");
             paramList.gen(out, context);
             out.print(")");
          }
       }
    }

    public void accept( Visitor v ) {
        v.visit(this);
    }

    private Expr receiver;
    private MethodInterface methodInterface;
    private ExprList paramList;

}
