package AST;

import java.lang.*;
import ws;
import java.io.*;
import SymbolTable;
import WhiteSpace;
import java.util.*;
import CompilerError.*;
import Interface.*;
import NameServer;
import GreenProject;

public class Method extends Identifier {

    public Method( String name, boolean isAbstract, int methodNumber ) {
       super(name);
       this.isAbstract = isAbstract;
       this.methodNumber = methodNumber;
       hasAssertAfter = false;    
       isRedefinition = false;
       localVariableList = new LocalVariableList();
       isInitMethod = false;
    }

    public String getJavaName() {
        if ( realJavaName == null ) {
           realJavaName = NameServer.getMethodJavaName( getName(), paramList );
        }
        return realJavaName;
    }

    public void setParamList(ParameterList paramList) {
       typeList = new TypeList();
       Enumeration e = paramList.elements();
       while ( e.hasMoreElements() ) 
          typeList.add( (Type ) ((Parameter ) e.nextElement()).getType() );
       this.paramList = paramList;
       searchName = TypeInterface.buildMethodSearchName( getName(), paramList );
    }

    public String getMethodInterfaceName() {
         if ( methodInterfaceName == null ) {
            String methodName;
            if ( isOperator )
               methodName = NameServer.getOperatorName(getName());
            else
               methodName = getName();
            methodInterfaceName = NameServer.getMethodInterfaceName(
                  methodName, returnType, paramList);
         }
         return methodInterfaceName;
    }

    public void genInterface( PrintWriter out, String className ) {
         out.println("interface " + getMethodInterfaceName() + " {");
         out.print("   ");
         if ( returnType != null ) {
             if ( returnType instanceof BasicType || 
                  GreenProject.subtypeEqualSubclass ||
                  returnType.getIsJavaClass()  ||
                  returnType.isArrayType()  
                  )
                out.print( returnType.getJavaName() + " " );
             else 
                out.print( "_Any " );
         }
         else
           out.print("void ");
         out.print( methodInterface.getJavaName() + "(" );
         paramList.gen(out);
         out.println(");");
         out.println("}");
         out.println("// " + className);
    }

    public String getClassMethodInfoNameVar() {
         if ( classMethodInfoNameVar == null ) {
            String name = getName();
            if ( isOperator )
               name = NameServer.getOperatorName(name);
            classMethodInfoNameVar = NameServer.getClassMethodInfoNameVar(
                  name, paramList);
         }
         return classMethodInfoNameVar;
    }

    public String getObjectMethodInfoNameVar() {
         if ( objectMethodInfoNameVar == null ) {
            String name = getName();
            if ( isOperator )
               name = NameServer.getOperatorName(name);
            objectMethodInfoNameVar = NameServer.getObjectMethodInfoNameVar(
                  name, paramList);
         }
         return objectMethodInfoNameVar;
    }

    public void gen( PrintWriter out, boolean isBasicType, Context context ) {
              
       context.setCurrentMethod(this);
       ws.add();
       if ( isInitMethod )
          ws.print( "public " );
       else
          ws.print( Method.getStringVisibility(visibility) + " " );
       if ( isBasicType )
          out.print("static ");
       if ( isAbstract && ! isBasicType )
         out.print("abstract ");
       if ( returnType != null ) {
          if ( returnType instanceof BasicType || 
               returnType.getIsJavaClass() ||
               GreenProject.subtypeEqualSubclass || 
               returnType.isArrayType()  
               )
             out.print( returnType.getJavaName() + " " );
          else 
             out.print( "_Any " );
       }
       else
         out.print("void ");
         
       if ( isOperator ) {
          out.print( NameServer.getOperatorName(getName()) + "(" );
       }
       else 
          out.print( getJavaName() + "(" );
       if ( isBasicType ) {
          out.print( context.getCurrentClass().getClassTypeInterface().getJavaName() + " _self" );
          if ( paramList.getSize() > 0 )
             out.print(", ");
       }
          
       paramList.gen(out);
       out.print(")");
       if ( isAbstract ) 
          out.println(";");
       else {
          out.println(" {");
          if ( context.getCurrentClass() != null &&
               context.getCurrentClass().getIsShellClass() )  {
             ws.add();
             String shell_ivcName = context.getCurrentClass().getShell_ivcName();
             ws.print( shell_ivcName + " shellV = (" + shell_ivcName + " ) " );
             if ( ((ClassTypeInterface) context
                 .getCurrentClass()
                    .getAttachedType().getTypeInterface())
                      .getIsReflective() ) 
                ws.println( "sv;" );
             else 
                ws.println( "ht.get(this);" );
             ws.sub();
          }
          if ( assertClause == null ) {
             localVariableList.gen(out);
             out.println();
             ws.add();
             if ( isOperator ) {
                if ( returnType != null ) 
                   ws.println( "return " + ((BasicType ) returnType).getJavaZeroValue() + ";");
             }
             else
                blockStatement.gen(out, context);
             ws.sub();
          }
          else { 
              if ( assertClause.getBeforeExpr() != null ) {
                  ws.add();
                  ws.print("if ( ! (");
                  assertClause.getBeforeExpr().gen(out, context);
                  out.println(") ) ");
                  ws.add();
                  ws.println("throw new PackedException( _AssertionBeforeException.$classObject._new$_MethodInfo(null) );");
                  ws.sub();
                  ws.sub();
              }
              
              if ( assertClause.getAfterExpr() != null && returnType != null ) {
                 ws.add();
                 ws.println( returnType.getJavaName() + " $return;" );
                 ws.println("$fim:");
                 ws.println("while (true) {");
                 ws.add();
              }
              if ( assertClause.getAssertVariableList() != null )
                 assertClause.getAssertVariableList().gen(out, context);
              
              localVariableList.gen(out);
              out.println();
              ws.add();
              blockStatement.gen(out, context);
              ws.sub();
              if ( assertClause.getAfterExpr() != null ) {
                  ws.sub();
                  
                  ws.add();
                  ws.print("if ( ! (");
                  assertClause.getAfterExpr().gen(out, context);
                  out.println(") ) ");
                  ws.add();
                  ws.println("throw new PackedException( _AssertionAfterException.$classObject._new$_MethodInfo(null) );");
                  ws.sub();
                  if ( returnType != null ) {
                     ws.println("}");  // while (true)
                     ws.sub();
                     ws.sub();
                  }
              }
          }
       ws.println("}");
       }
       ws.sub();
    }

    public void genCaseCallMethod(PrintWriter out) {
         ws.println("case " + methodNumber + " :");
         ws.add();
         ws.println("{");
         ws.add();
         ws.println("if ( args._getSize() != " + paramList.getSize() + ")");
         ws.add();
         ws.println("throw new PackedException( _WrongParametersException.$classObject._new() );");
         ws.sub();
         Enumeration e;
         if ( paramList.getSize() > 0 ) {
            e = paramList.elements();
            while ( e.hasMoreElements() ) {
               Parameter p = (Parameter ) e.nextElement();
               ws.println( p.getType().getJavaName() + " " + p.getParameterInfoVarName() + ";");
            }
            ws.println("try {");
            ws.add();
            int paramNumber = 0;
            e = paramList.elements();
            while ( e.hasMoreElements() ) {
               Parameter p = (Parameter ) e.nextElement();
               
               ws.print( p.getParameterInfoVarName() + " = (" );
               if ( p.getType() instanceof BasicType ) {
                  BasicType basicType = (BasicType ) p.getType();
                  out.println( "(" + basicType.getJavaWrapperName() + " ) args.v[" +
                     paramNumber + "])._get();");
               }
               else
                  if ( p.getType().getName().compareTo("String") == 0 ) {
                     out.println("(_PackedString ) args.v[" + paramNumber + "])._getString();");
                  }
                  else
                     out.println( p.getType().getJavaName() + " ) args.v[" + 
                        paramNumber + "];");
               paramNumber++;
            }
            
            ws.sub();
            ws.println("} catch ( ClassCastException e ) {"); // try      
            ws.add();
            ws.println("throw new PackedException( _WrongParametersException.$classObject._new() );");
            ws.sub();
            ws.println("}");
         }
         ws.println("try {");
         ws.add();
         /*
         any = _Integer.$classObject._new$int( _n(1, 2) );
         any = null;  _n(1, 2);
         any = _n(1, 2);
         */
         ws.print("any = ");
         Type returnType = methodInterface.getReturnType();
            // A Java class cannot be assigned to a variable of type _Any 
         if ( returnType != null && returnType.getIsJavaClass() )
            returnType = null;
            
         if ( returnType == null )
            out.println("null;");
         else {
            if ( returnType instanceof BasicType ) {
               BasicType basicType = (BasicType ) returnType;
               out.print( basicType.getNewWrapperString() );  // _Integer.$classObject._new$int(
            }
         }
         if ( returnType == null ) 
            ws.print( methodInterface.getJavaName() + "(");
         else 
            out.print( methodInterface.getJavaName() + "(");
            
         e = paramList.elements();
         int size = paramList.getSize();
         while ( e.hasMoreElements() ) {
            out.print( ((Parameter ) e.nextElement()).getParameterInfoVarName() );
            if ( --size > 0 )
               out.print(", ");
         }
         out.print(")");
         if ( returnType == null )
            out.println(";");
         else
            if ( returnType instanceof BasicType ) 
               out.println(");");
            else
               out.println(";");
         
         ws.sub();
         ws.println("} catch ( Exception e ) {");
         ws.add();
         ws.println("throw new PackedException( e );");
         ws.sub();
         ws.println("}");
         ws.sub();
         ws.println("}");
         ws.println("break;");
         ws.sub();
    }

    public void genObjectMethodInfo(PrintWriter out) {
         ws.println("{");
         ws.add();
         // paramList.genDeclarationParameter(out);
         Enumeration e = paramList.elements();
         while ( e.hasMoreElements() ) {
            Parameter p = (Parameter ) e.nextElement();
            ws.println("_ParameterInfo " + p.getParameterInfoVarName() + 
                " = (_ParameterInfo ) _ParameterInfo.$classObject._new$String$_ClassInfo$_MethodInfo$$Array$1$_CodeAnnotation$boolean(\"" + p.getName() + "\", " + 
                ((ClassTypeInterface ) p.getType().getTypeInterface()).getAccessToClassInfo()
                    + ", null, null, false);" );
         }
         ws.println( getObjectMethodInfoNameVar() + " = (_ObjectMethodInfo ) _ObjectMethodInfo.$classObject._new$String$int$_MethodBodyInfo$int$$Array$1$_ClassInfo$_ClassInfo$$Array$1$_ParameterInfo$_ClassInfo$$Array$1$_CodeAnnotation(");
         ws.add();
         ws.println("\"" + getName() + "\",");
         ws.println( methodNumber + "," );
         ws.println( "null," );
         ws.println( Method.getStringAccessVisibility(visibility) + "," );
         
         ws.println("new " + 
             TypeInterface.buildArrayName("_ClassInfo", 1) + "(");
         ws.add();
         ws.println( "new _ClassInfo[] { ");
         ws.add();
         e = paramList.elements();
         int size = paramList.getSize();
         while ( e.hasMoreElements() ) {
            Parameter p = (Parameter ) e.nextElement();
            ws.print(  ((ClassTypeInterface ) p.getType().getTypeInterface()).getAccessToClassInfo() );
            if ( --size > 0 )
               out.println(",");
            else
               out.println();
         }
         ws.println("}");
         ws.sub();
         ws.println("),");
         ws.sub();
         if ( returnType == null )
            ws.println("null,");
         else
            ws.println( ( (ClassTypeInterface ) returnType.getTypeInterface()).getAccessToClassInfo()
                + ","  );
         ws.println("new " + 
             TypeInterface.buildArrayName("_ParameterInfo", 1) + "(");
         ws.add();
         ws.println("new _ParameterInfo[] {");
         ws.add();
         e = paramList.elements();
         size = paramList.getSize();
         while ( e.hasMoreElements() ) {
            Parameter p = (Parameter ) e.nextElement();
            ws.print(  p.getParameterInfoVarName() );
            if ( --size > 0 )
               out.println(",");
            else
               out.println();
         }
         ws.println("}");
         ws.sub();
         ws.println("),");
         ws.sub();
         if ( exceptionType == null ) 
            ws.println("null,");
         else 
            ws.println( ( (ClassTypeInterface ) exceptionType.getTypeInterface()).
                    getClassObjectInterface().getAccessToClassInfo() + ","  );
         ws.println("null");
         ws.println(");");
         
         ws.sub();
         e = paramList.elements();
         while ( e.hasMoreElements() ) {
            Parameter p = (Parameter ) e.nextElement();
            ws.println( p.getParameterInfoVarName() + "._$setDeclaringMethod$_MethodInfo(" +
               getObjectMethodInfoNameVar() + ");" );
         }
         
         ws.sub();
         ws.println("}");
    }

    public void genClassMethodInfo(PrintWriter out) {
         ws.println("{");
         ws.add();
         // paramList.genDeclarationParameter(out);
         Enumeration e = paramList.elements();
         while ( e.hasMoreElements() ) {
            Parameter p = (Parameter ) e.nextElement();
            ws.println("_ParameterInfo " + p.getParameterInfoVarName() + 
                " = (_ParameterInfo ) _ParameterInfo.$classObject._new$String$_ClassInfo$_MethodInfo$$Array$1$_CodeAnnotation$boolean(\"" + p.getName() + "\", " + 
                ((ClassTypeInterface ) p.getType().getTypeInterface()).getAccessToClassInfo()
                + ", null, null, false);" );
         }
         ws.println( getClassMethodInfoNameVar() + " = (_ClassMethodInfo ) _ClassMethodInfo.$classObject._new$String$boolean$int$_MethodBodyInfo$int$$Array$1$_ClassInfo$_ClassInfo$$Array$1$_ParameterInfo$_ClassInfo$$Array$1$_CodeAnnotation(");
         ws.add();
         
         ws.println("\"" + getName() + "\",");
         ws.println( isAbstract ? "true," : "false," );
         ws.println( methodNumber + "," );
         ws.println( "null," );
         ws.println( Method.getStringAccessVisibility(visibility) + "," );
         
         ws.println( "new " + TypeInterface.buildArrayName("_ClassInfo", 1) + "(" );
         ws.add();
         ws.println( "new _ClassInfo[] { ");
         ws.add();
         e = paramList.elements();
         int size = paramList.getSize();
         while ( e.hasMoreElements() ) {
            Parameter p = (Parameter ) e.nextElement();
            ws.print(  ((ClassTypeInterface ) p.getType().getTypeInterface()).getAccessToClassInfo() );
            if ( --size > 0 )
               out.println(",");
            else
               out.println();
         }
         ws.println("}");
         ws.sub();
         ws.println("),");
         ws.sub();
         
         
         if ( returnType == null )
            ws.println("null,");
         else
            ws.println( ((ClassTypeInterface ) returnType.getTypeInterface()).getAccessToClassInfo()
                + ","  );
         ws.println( "new " + TypeInterface.buildArrayName("_ParameterInfo", 1) + "(" );
         ws.add();
         ws.println("new _ParameterInfo[] {");
         ws.add();
         e = paramList.elements();
         size = paramList.getSize();
         while ( e.hasMoreElements() ) {
            Parameter p = (Parameter ) e.nextElement();
            ws.print(  p.getParameterInfoVarName() );
            if ( --size > 0 )
               out.println(",");
            else
               out.println();
         }
         ws.println("}");
         ws.sub();
         ws.println("),");
         ws.sub();
         if ( exceptionType == null ) 
            ws.println("null,");
         else 
            ws.println( ( (ClassTypeInterface ) exceptionType.getTypeInterface()).
                    getClassObjectInterface().getAccessToClassInfo() + ","  );
         ws.println("null");
         ws.println(");");
         
         ws.sub();
         e = paramList.elements();
         while ( e.hasMoreElements() ) {
            Parameter p = (Parameter ) e.nextElement();
            ws.println( p.getParameterInfoVarName() + "._$setDeclaringMethod$_MethodInfo(" +
               getObjectMethodInfoNameVar() + ");" );
         }
         
         ws.sub();
         ws.println("}");
    }

    static public final int public_v = 0, private_v = 1, subclass_v = 2;
    
    static String getStringAccessVisibility( int theVisibility ) {
     switch ( theVisibility ) {
        case Method.public_v :
           return "$co$MethodInfo._public_v";
        case Method.private_v :
           return "$co$MethodInfo._private_v";
        case Method.subclass_v :
           return "$co$MethodInfo._subclass_v";
        default :
           return "Internal error: wrong visibility";
     }
    }
    static String getStringVisibility( int theVisibility ) {
     switch ( theVisibility ) {
        case Method.public_v :
           return "public";
        case Method.private_v :
           return "private";
        case Method.subclass_v :
           return "protected";
        default :
           return "Internal error: wrong visibility";
     }
    }
    
    

    public void accept( Visitor v ) {
        v.visit(this);
    }

    public MethodInterface getMethodInterface() {
       return methodInterface;
    }

    public void setMethodInterface( MethodInterface methodInterface ) {
       this.methodInterface = methodInterface;
    }

    public boolean getHasAssertAfter() {
       return hasAssertAfter;
    }

    public void setHasAssertAfter( boolean hasAssertAfter ) {
       this.hasAssertAfter = hasAssertAfter;
    }

    public boolean getIsRedefinition() {
       return isRedefinition;
    }

    public void setIsRedefinition( boolean isRedefinition ) {
       this.isRedefinition = isRedefinition;
    }

    public boolean getIsOperator() {
       return isOperator;
    }

    public void setIsOperator( boolean isOperator ) {
       this.isOperator = isOperator;
    }

    public TypeList getTypeList() {
       return typeList;
    }

    public void setTypeList( TypeList typeList ) {
       this.typeList = typeList;
    }

    public boolean getIsAbstract() {
       return isAbstract;
    }

    public ParameterList getParamList() {
       return paramList;
    }

    public Type getExceptionType() {
       return exceptionType;
    }

    public void setExceptionType( Type exceptionType ) {
       this.exceptionType = exceptionType;
    }

    public Type getReturnType() {
       return returnType;
    }

    public void setReturnType( Type returnType ) {
       this.returnType = returnType;
    }

    public void setAssertClause( AssertClause assertClause ) {
       this.assertClause = assertClause;
    }

    public LocalVariableList getLocalVariableList() {
       return localVariableList;
    }

    public void setLocalVariableList( LocalVariableList localVariableList ) {
       this.localVariableList = localVariableList;
    }

    public BlockStatement getBlockStatement() {
       return blockStatement;
    }

    public void setBlockStatement( BlockStatement blockStatement ) {
       this.blockStatement = blockStatement;
    }

    public int getMethodNumber() {
       return methodNumber;
    }

    public int getVisibility() {
       return visibility;
    }

    public void setVisibility( int visibility ) {
       this.visibility = visibility;
    }

    public String getSearchName() {
       return searchName;
    }

    public boolean getIsInitMethod() {
       return isInitMethod;
    }

    public void setIsInitMethod( boolean isInitMethod ) {
       this.isInitMethod = isInitMethod;
    }

    private MethodInterface methodInterface;
    private boolean hasAssertAfter;
    private boolean isRedefinition;
    private String realJavaName;
    private String methodInterfaceName;
    private String objectMethodInfoNameVar;
    private String classMethodInfoNameVar;
    private boolean isOperator;
    private TypeList typeList;
    private boolean isAbstract;
    private ParameterList paramList;
    private Type exceptionType;
    private Type returnType;
    private AssertClause assertClause;
    private LocalVariableList localVariableList;
    private BlockStatement blockStatement;
    private int methodNumber;
    private int visibility;
    private String searchName;
    private boolean isInitMethod;

}
