package AST;

import java.lang.*;
import ws;
import SymbolTable;
import WhiteSpace;
import java.util.*;
import CompilerError.*;
import Interface.*;

abstract public class Type extends Identifier {

    public Type() {
       super();
       isClassObjectType = false;
    }

    public Type(String name) {
       super(name);
       isClassObjectType = false;
    }

    public String getArrayName(int numDimensions) {
         return TypeInterface.buildArrayName(this, numDimensions);
    }

    public String getJavaZeroValue() {
         return "null";
    }

    public boolean getIsJavaClass() {
         return typeInterface.getIsJavaClass();
    }

    public boolean isArrayType() {
         return name.indexOf("$Array$") >= 0;
    }

    public String getInterfaceName() {
         return name;
    }

    public boolean isSupertypeOf(Type subtype) {
         return typeInterface.isSupertypeOf(subtype.getTypeInterface());
    }

    public void reify( TypeInterfaceList til, ErrorMsg em ) {
       
         if ( (typeInterface = til.get( getInterfaceName() )) == null ) 
            em.show("Class " + getInterfaceName() + " not found");
    }

    public boolean hasArray() {
        return false;
    }

    abstract public void accept(Visitor v);

    public static VoidType voidType = new VoidType();
    public static LiteralNilExpr nilExpr = new LiteralNilExpr();
    public static NilType nilType = new NilType();
    public static BooleanType booleanType = new BooleanType();
    public static ByteType    byteType    = new ByteType();
    public static IntegerType integerType = new IntegerType();
    public static LongType    longType    = new LongType();
    public static CharType    charType    = new CharType();
    public static RealType    realType    = new RealType();
    public static DoubleType  doubleType  = new DoubleType();
    public static UndefinedType undefinedType = new UndefinedType();
    public static StringType  stringType  = new StringType("String");
    public static ClassTypeInterface anyTypeInterface, anyClassTypeInterface, 
      reflectiveAnyClassTypeInterface,
      anyClassObjectTypeInterface, anyValueTypeInterface, nilTypeInterface,
      CharTypeInterface, BooleanTypeInterface, ByteTypeInterface, 
      IntegerTypeInterface, LongTypeInterface, RealTypeInterface, DoubleTypeInterface,
      StringTypeInterface,
      wrapperCharTypeInterface,
      wrapperByteTypeInterface,
      wrapperBooleanTypeInterface,
      wrapperIntegerTypeInterface,
      wrapperLongTypeInterface,
      wrapperRealTypeInterface,
      wrapperDoubleTypeInterface;
      
    
    public static ClassType 
     anyType, anyClassType, anyValueType, 
     anyClassObjectType, reflectiveAnyClassType,
     wrapperCharType, wrapperByteType, wrapperBooleanType,
     wrapperIntegerType, wrapperLongType, wrapperRealType,
     wrapperDoubleType;
    public static Type []arrayAnyValue = {
      charType, byteType, booleanType,
      doubleType, integerType, longType, realType, 
      stringType };
    
    private static Hashtable wrapperTable;
    static {
    anyType = new ClassType("Any", null);
    anyType.setRealJavaName();
    anyClassType = new ClassType("AnyClass", null);
    anyClassType.setRealJavaName();
    anyValueType = new ClassType("AnyValue", null);
    anyValueType.setRealJavaName();
    anyClassObjectType = new ClassType("AnyClassObject", null);
    anyClassObjectType.setRealJavaName();
    reflectiveAnyClassType = new ClassType("ReflectiveAnyClass", null);
    wrapperCharType = new ClassType("Char", null);
    wrapperByteType = new ClassType("Byte", null); 
    wrapperBooleanType = new ClassType("Boolean", null);
    wrapperIntegerType = new ClassType("Integer", null); 
    wrapperLongType = new ClassType("Long", null);
    wrapperRealType = new ClassType("Real", null);
    wrapperDoubleType = new ClassType("Double", null);
    
    wrapperTable = new Hashtable();
    wrapperTable.put("Char", "char");
    wrapperTable.put("Boolean", "boolean");
    wrapperTable.put("Byte", "byte");
    wrapperTable.put("Integer", "integer");
    wrapperTable.put("Long", "long");
    wrapperTable.put("Real", "real");
    wrapperTable.put("Double", "double");
    }

    public TypeInterface getTypeInterface() {
       return typeInterface;
    }

    public void setTypeInterface( TypeInterface typeInterface ) {
       this.typeInterface = typeInterface;
    }

    public boolean getIsClassObjectType() {
       return isClassObjectType;
    }

    public void setIsClassObjectType( boolean isClassObjectType ) {
       this.isClassObjectType = isClassObjectType;
    }

    private TypeInterface typeInterface;
    private boolean isClassObjectType;

}
