import java_cup.runtime.*;
import CompilerError.*;
import Project.*;
import java.io.*;

public class gc {

        
    public static char []readFile( String fileName ) {
        FileReader stream;
        int numChRead;
        
        File file = new File(fileName);
        if ( ! file.exists() || ! file.canRead() ) {
           System.out.println("Either the file " + fileName + " does not exist or it cannot be read");
           return null;
         }
         try { 
             stream = new FileReader(file);  
         } catch ( FileNotFoundException e ) {
            System.out.println("Something wrong: file " + fileName + " does not exist anymore");
            return null;
         }
                // one more character for '\0' at the end that will be added by the
                // compiler
         char []input = new char[ (int ) file.length() + 1 ];
            
         try {
            numChRead = stream.read( input, 0, (int ) file.length() );
         } catch ( IOException e ) {
            System.out.println("Error reading file " + fileName);
            return null;
         }
                
         if ( numChRead != file.length() ) {
            System.out.println("Read error in file " + fileName);
            return null;
         }
         input[numChRead] = '\0';
         try {
            stream.close();
         } catch ( IOException e ) {
            System.out.println("Error in handling the file " + fileName );
            return null;
         }
         return input;
    }


    public static void main( String []args ) {
        
        PrintWriter out;
        FileOutputStream fileOut;
        ErrorMsg errorMsg = new ErrorMsg( new PrintWriter(System.out) );
        
        if ( args.length > 2  )  {
            System.out.println("Usage:\n   gc input ");
            System.out.println("input is the project file");
        }
        else {
            String filename = args[0];
            try { 
               java.io.InputStream inp = new java.io.FileInputStream(filename);
            } catch ( Exception e ) {
                errorMsg.show("unable to open " + filename);
                return;
            }
            char []input = readFile(filename);
            MyLexer lexer = new MyLexer( 
               errorMsg,
               input );
            Sym token;
            if ( args.length == 2 ) 
                try { 
                   fileOut = new FileOutputStream(args[1]);
                   out = new PrintWriter(fileOut);
                } catch ( Exception e ) {
                    errorMsg.show("Unable to open " + args[1]);
                    return;
                }
            else {
                out = new PrintWriter(System.out);
                fileOut = null;
            }
      
      try {
        while ( lexer.token != Sym.EOF ) {
            switch ( lexer.token ) {
                case Sym.STRING : 
                    out.println( "string(" + (String ) lexer.stringValue + ")" );
                    break;
                case Sym.IDENT :
                    out.println( "IDENT(" + (String ) lexer.ident + ")" );
                    break;
                case Sym.INTEGERCONST : 
                    out.println( "Integerconst = " + lexer.stringValue );
                    break;
                case Sym.BYTECONST :
                    out.println( "ByteConst = " + lexer.stringValue );
                    break;
                case Sym.CHARCONST :
                    out.println( "CharConst = " + lexer.stringValue );
                    break;
                case Sym.REALCONST :
                    out.println( "Real = " + lexer.stringValue );
                    break;
                case Sym.DOUBLECONST :
                    out.println( "Double = " + lexer.stringValue );
                    break;
                default :
                    out.println( symarray[lexer.token] );
                    
            }
         lexer.next();
        }
      } catch ( Exception e ) {
        System.out.println("Error in lexical analyses " + e);
      }
    out.println("This is the end");
    out.flush();
    if ( fileOut != null ) 
       try {
          fileOut.close();
       } catch ( Exception e) { }
        }
    }
            
  static String symarray[] = new String[150];
  static {
     for (int i = 0; i < 150; i++)
        symarray[i] = "Illegal Symbol";
symarray[Sym.JAVABEGIN] = "java$begin";
symarray[Sym.JAVAEND] = "java$end";
symarray[Sym.JAVAUSE] = "java$use";
symarray[Sym.EOF] = "EOF";
symarray[Sym.CHARCONST] = "CharConst";
symarray[Sym.INTEGERCONST] = "IntegerConst";
symarray[Sym.BYTECONST] = "ByteConst";
symarray[Sym.LONGCONST] = "LongConst";
symarray[Sym.REALCONST] = "RealConst";
symarray[Sym.DOUBLECONST] = "DoubleConst";
symarray[Sym.STRING] = "String";

symarray[Sym.ABSTRACT] = "abstract";
symarray[Sym.AFTER] = "after";
symarray[Sym.AND] = "and";
symarray[Sym.ARRAY] = "array";
symarray[Sym.ASSERT] = "assert";
symarray[Sym.BEFORE] = "before";
symarray[Sym.BEGIN] = "begin";
symarray[Sym.BOOLEAN] = "boolean";
symarray[Sym.BREAK] = "break";
symarray[Sym.BYTE] = "byte";
symarray[Sym.CASE] = "case";
symarray[Sym.CHAR] = "char";
symarray[Sym.CLASS] = "class";
symarray[Sym.CONST] = "const";
symarray[Sym.DO] = "do";
symarray[Sym.DOUBLE] = "double";
symarray[Sym.ELSE] = "else";
symarray[Sym.END] = "end";
symarray[Sym.ENDIF] = "endif";
symarray[Sym.ENUM] = "enum";
symarray[Sym.EXCEPTION] = "exception";
symarray[Sym.FALSE] = "false";
symarray[Sym.FOR] = "for";
symarray[Sym.IF] = "if";
symarray[Sym.INTEGER] = "integer";
symarray[Sym.LONG] = "long";
symarray[Sym.LOOP] = "loop";
symarray[Sym.NIL] = "nil";
symarray[Sym.NOT] = "not";
symarray[Sym.OBJECT] = "object";
symarray[Sym.OR] = "or";
symarray[Sym.OTHERWISE] = "otherwise";
symarray[Sym.PROC] = "proc";
symarray[Sym.PRIVATE] = "private";
symarray[Sym.PUBLIC] = "public";
symarray[Sym.REAL] = "real";
symarray[Sym.REFLECTIVE] = "reflective";
symarray[Sym.REPEAT] = "repeat";
symarray[Sym.RESULT] = "result";
symarray[Sym.RETURN] = "return";
symarray[Sym.SELF] = "self";
symarray[Sym.SHELL] = "shell";
symarray[Sym.STRING] = "string";
symarray[Sym.SUBCLASS] = "subclass";
symarray[Sym.SUBCLASSOF] = "subclassof";
symarray[Sym.SUPER] = "super";
symarray[Sym.THEN] = "then";
symarray[Sym.TO] = "to";
symarray[Sym.TRUE] = "true";
symarray[Sym.TRY] = "try";
symarray[Sym.TYPE] = "type";
symarray[Sym.UNTIL] = "until";
symarray[Sym.VAR] = "var";
symarray[Sym.WHILE] = "while";
symarray[Sym.XOR] = "xor";
symarray[Sym.IDENT] = "IDENT";


symarray[Sym.AT] = "@";
symarray[Sym.ASSIGN] = "=";
symarray[Sym.BINAND] = "&";
symarray[Sym.BINNOT] = "~";
symarray[Sym.BINOR] = "|";
symarray[Sym.BINXOR] = "^";
symarray[Sym.RIGHTSHIFT] = ">>";
symarray[Sym.LEFTSHIFT] = "<<";
symarray[Sym.CANCELA] = "#";
symarray[Sym.COLON] = ":";
symarray[Sym.COMMA] = ",";
symarray[Sym.PERIOD] = ".";
symarray[Sym.SEMICOLON] = ";";
symarray[Sym.EQ] = "==";
symarray[Sym.NEQ] = "<>";
symarray[Sym.GE] = ">=";
symarray[Sym.GT] = ">";
symarray[Sym.LE] = "<=";
symarray[Sym.LT] = "<";
symarray[Sym.PLUS] = "+";
symarray[Sym.MINUS] = "-";
symarray[Sym.MULT] = "*";
symarray[Sym.DIV] = "/";
symarray[Sym.REMAINDER] = "%";
symarray[Sym.PLUSPLUS] = "++";
symarray[Sym.MINUSMINUS] = "--";
symarray[Sym.LEFTPAR] = "(";
symarray[Sym.RIGHTPAR] = ")";
symarray[Sym.LEFTSB] = "[";
symarray[Sym.RIGHTSB] = "]";
symarray[Sym.THREEPERIOD] = "...";

     
   }

}


